!
! Here's a rendering API for the PowerVR2:
!
! The PowerVR2 wants the polygon data sorted on type:
!  opaque, translucent, opaque modifier, translucent modifier, punchthrough.
! The types may come in any order, but all polygons of one type must
!  be sent before polygons of another type are sent.

! The rendering API allows you to either send polygons directly to the TA,
!  in which case you must honor the above rule, or to create vertex buffers
!  in system memory, where polygons are gathered until the current render
!  is done.
! System-memory vertex buffers run slower but allow you to send polygons
!  in any order.

! The current implementation is mainly interrupt driven. Rendering
!  is normally 1 frame after the vertex feeding.
! The current implementation does not use DMA for any data transfers.


! Here's the API, sortof:
!
! rdInit()
! rdSetupNativeBuffers()
! rdSetupVertexBuffer()  for each system-memory vertex buffer you want
!
! for each frame:
!   rdBeginScene(type)
!
!   If type=TRUE, then you're on your own. Feed your polygon data to the TA,
!    including End-Of-List when you are finished with each enabled polygon type.
!   If type=FALSE, there is much more though:
!
!     rdSetupDirectPath() to say that the specified path should go straight
!	to the TA, not to system memory.
!	After this, writes using the write-ptrs in the API will go straight
!	to the TA.
!
!     rdFinishDirectPath() closes the currently-DirectPathed type and flushes
!	any associated vertexbuffer.
!
!     rdFlushVertexBuffer() can be done to flush the vertexbuffer of the
!	currently-DirectPathed type. If there is no vertexbuffer, it's a
!	no-op.
!
!     rdDrawPrimitives() can be used to send primitives of the various types,
!	either to vertexbuffers or straight to TA (if a DirectPath has been
!	set up).
!
!     Alternatively, one can peek in the innards of the renderer and write
!     data directly to the TA...
!
!   rdEndScene()
!   rdFlip()
!
!
! rdShutdown()
!
!
! Also, one can create callbacks on most of the TA/3D interrupts.
!

		.INCLUDE "PVR2_Regs.i"
		.INCLUDE "Screen.i"
		.INCLUDE "Render.i"
	
		.SECTION .text



!------------------------------------------------------------------------
! void rdInit(void)
!
! Initializes TA + 3D rendering unit
!

		.GLOBL	_rdInit
		.ALIGN	5
_rdInit:
		sts.l	pr,@-r15

		mov.l	_rdInit_PVR2Regs,r7

		mov	#3,r0
		mov.l	r0,@(COREDISABLE*4,r7)	! Reset TA + 3D
		mov	#0,r0
		mov.l	r0,@(COREDISABLE*4,r7)	! Un-reset TA + 3D

		mov.l	_rdInit_RegsToSetup,r0	! Setup default init-values for various TA/3D regs
_rdInit_setReg:
		mov.l	@r0+,r1
		mov	#-1,r2
		cmp/eq	r1,r2
		bt	_rdInit_setRegsDone
		mov.l	@r0+,r2
		bra	_rdInit_setReg
		mov.l	r2,@r1
_rdInit_setRegsDone:

		
		mov.l	_rdInit_stdClearMem,r0
		mov.l	_rdInit_VertexBuffers,r4
		mov	#VertexBuffer_SIZEOF*5,r5
		jsr	@r0
		extu.b	r5,r5

		mov.l	_rdInit_stdClearMem,r0
		mov.l	_rdInit_Render,r4
		mov	#Render_SIZEOF,r5
		jsr	@r0
		extu.b	r5,r5

		bsr	_rdSetRenderDoneUserCallback
		mov	#0,r4
		bsr	_rdSetVSyncUserCallback
		mov	#0,r4
		bsr	_rdSetScanlineInt1UserCallback
		mov	#0,r4
		bsr	_rdSetScanlineInt2UserCallback
		mov	#0,r4
		
		mov	#0x32,r4
		mov.l	_rdInit_SetExceptionEventHandler,r0
		shll2	r4
		mov.l	_rdInit_rdIRQ9Handler,r5
		jsr	@r0
		shll2	r4

		mov.l	_rdInit_IntBase,r0
		mov	#-1,r1
		mov.l	r1,@(0,r0)
		mov.l	r1,@(8,r0)
		mov.l	_rdInit_EnableMask1,r1
		mov.l	_rdInit_EnableMask2,r2
		mov.l	r1,@(0x30,r0)
		mov.l	r2,@(0x38,r0)

		lds.l	@r15+,pr
		rts
		nop

				.ALIGN	2
_rdInit_IntBase:		.LONG	0xa05f6900
_rdInit_PVR2Regs:		.LONG	PVR2REGS
_rdInit_RegsToSetup:		.LONG	RegsToSetup
_rdInit_stdClearMem:		.LONG	_stdClearMem
_rdInit_VertexBuffers:		.LONG	_VertexBuffers
_rdInit_Render:			.LONG	_Render
_rdInit_SetExceptionEventHandler: .LONG	_SetExceptionEventHandler
_rdInit_rdIRQ9Handler:		.LONG	rdIRQ9Handler
_rdInit_EnableMask1:		.LONG	0x0000003c | 0xffdff840
_rdInit_EnableMask2:		.LONG	0x0000000d | 0xfffffff2



!------------------------------------------------------------------------
! void rdShutdown(void)
!
!

		.GLOBL	_rdShutdown
		.ALIGN	5
_rdShutdown:
		sts.l	pr,@-r15
	
		bsr	_rdWaitTAIdle
		nop

		bsr	_rdWaitRenderIdle
		nop

		lds.l	@r15+,pr
	
		mov.l	_rdShutdown_IntBase,r1
		mov	#0,r0
		mov.l	r0,@(0x30,r1)			! Disable all ints
		mov.l	r0,@(0x38,r1)
		mov	#-1,r0
		mov.l	r0,@(0,r1)
		rts
		mov.l	r0,@(8,r1)

			.ALIGN	2
_rdShutdown_IntBase:	.LONG	0xa05f6900



!------------------------------------------------------------------------
! void rdSetupNativeBuffers(int opaqueSize, int opaqueModSize, int translucentSize,
!	int translucentModSize, int punchthroughSize, int nativeBufferSize)
!
! Sets up buffers
!
! in	r4		int opaqueSize
!	r5		int opaqueModSize
!	r6		int translucentSize
!	r7		int translucentModSize
!	@(0,r15)	int punchthroughSize
!	@(4,r15)	int nativeBufferSize
!

		.GLOBL	_rdSetupNativeBuffers
		.ALIGN	5
_rdSetupNativeBuffers:
		sts.l	pr,@-r15
		mov.l	_rdSetupNativeBuffers_Render,r0
		mov	r0,r3

		mov.l	@(4,r15),r0
		mov.l	@(8,r15),r1
		mov.l	r1,@(Render_NativeBufferSize,r3)

		tst	r0,r0
		movt	r1
		tst	r7,r7
		addc	r1,r1
		tst	r6,r6
		addc	r1,r1
		tst	r5,r5
		addc	r1,r1
		tst	r4,r4
		addc	r1,r1
		mov	#0x1f,r2
		xor	r2,r1
		mov.l	r1,@(Render_EnabledTypes,r3)

		mov	#4,r1
		shld	r1,r0
		or	r7,r0
		shld	r1,r0
		or	r6,r0
		shld	r1,r0
		or	r5,r0
		shld	r1,r0
		or	r4,r0
		mov.l	r0,@(Render_BucketSizes,r3)

		mov.l	_rdSetupNativeBuffers_Screen,r2

		mov.l	@(Screen_Width,r2),r4
		mov.l	@(Screen_Height,r2),r5
		mov.l	r4,@(Render_TargetWidth,r3)
		mov.l	r5,@(Render_TargetHeight,r3)
		mova	_rdSetupNativeBuffers_FormatTranslation,r0
		mov.l	@(Screen_Format,r2),r6
		mov.b	@(r0,r6),r6
		mov.l	r6,@(Render_TargetFormat,r3)

		mova	_rdSetupNativeBuffers_BPPShift,r0
		mov.b	@(r0,r6),r6
		shld	r6,r4
		mov.l	r4,@(Render_TargetPitch,r3)

		mov.l	@(Render_TargetWidth,r3),r4
		mov	#-5,r6
		mov.l	@(Render_TargetHeight,r3),r5
		mov	#0x1f,r0
		add	r0,r4
		add	r0,r5
		shld	r6,r4
		shld	r6,r5
		mov.l	r4,@(Render_TilesX,r3)
		mov.l	r5,@(Render_TilesY,r3)

		mov	r3,r7
		add	#Render_State,r7
		mov	#0,r0
		mov.b	r0,@(RenderState_CurrentTABufferIdx,r7)
		mov.b	r0,@(RenderState_Current3DBufferIdx,r7)
		mov	#1,r0
		mov.b	r0,@(RenderState_Current2DBufferIdx,r7)

		mov.l	_rdSetupNativeBuffers_RenderBuffer0,r0
		mov.l	_rdSetupNativeBuffers_RenderBuffer1,r1
		mov.l	r0,@(RenderState_CurrentTABuffer,r7)
		mov.l	r0,@(RenderState_Current3DBuffer,r7)
		mov.l	r1,@(RenderState_Current2DBuffer,r7)
	
		mov.l	_rdSetupNativeBuffers_RenderBuffer0,r0
		mov	r0,r4
		bsr	rdInitRenderBuffer
		mov	#0,r5

		mov.l	_rdSetupNativeBuffers_RenderBuffer1,r0
		mov	r0,r4
		mov	#0x40,r5
		bsr	rdInitRenderBuffer
		shll16	r5

		mov.l	_rdSetupNativeBuffers_Render,r0
		mov.l	_rdSetupNativeBuffers_scSetBuffer,r2
		mov.l	@(Render_State+RenderState_Current2DBuffer,r0),r1
		mov.l	@(Render_TargetPitch,r0),r5
		jsr	@r2
		mov.l	@(RenderBuffer_ScreenBase,r1),r4

		lds.l	@r15+,pr
		rts
		nop


						.ALIGN	5
_rdSetupNativeBuffers_Screen:			.LONG	_Screen
_rdSetupNativeBuffers_Render:			.LONG	_Render
_rdSetupNativeBuffers_RenderBuffer0:		.LONG	RenderBuffer0
_rdSetupNativeBuffers_RenderBuffer1:		.LONG	RenderBuffer1
_rdSetupNativeBuffers_scSetBuffer:		.LONG	_scSetBuffer
						.ALIGN	3
_rdSetupNativeBuffers_FormatTranslation:	.BYTE	RENDER_ARGB1555
						.BYTE	RENDER_RGB565
						.BYTE	RENDER_RGB888
						.BYTE	RENDER_ARGB8888
						.ALIGN	2
_rdSetupNativeBuffers_BPPShift:			.BYTE	1, 1, 1, 1, 0, 2, 2, 1



!------------------------------------------------------------------------
! void rdInitRenderBuffer(RenderBuffer* buffer, int baseAddr)
!
! Initializes one rendering buffer; builds Tile Array and computes
! memory addresses
!

		.ALIGN	5
rdInitRenderBuffer:
		mov.l	r10,@-r15
		mov.l	r9,@-r15
		mov.l	r8,@-r15

		add	#-5*16,r15

		mov.l	rdInitRenderBuffer_Render,r0
		mov	r0,r8
		mov.l	r5,@(RenderBuffer_ScreenBase,r4)
		mov.l	@(Render_TargetPitch,r8),r0		! Reserve space for framebuffer
		mov.l	@(Render_TargetHeight,r8),r1
		mul.l	r0,r1
		sts	macl,r0
		add	r0,r5
		mov.l	r5,@(RenderBuffer_TileArrayBase,r4)

		mov.l	@(Render_TilesX,r8),r9			! Reserve space for Tile Array
		mov.l	@(Render_TilesY,r8),r10
		mul.l	r9,r10
		sts	macl,r0
		mov	r0,r1
		add	r0,r0
		add	r1,r0
		shll2	r0
		shll	r0
		add	#6*4,r0					! Space for init-tile too, please
		add	r0,r5

		mov	#2,r0					! Some scratch space, in case of overflow later
		shll8	r0
		add	r0,r5
		mov.l	r5,@(RenderBuffer_PrimAllocStart,r4)
		mov.l	r5,@(RenderBuffer_PPAllocStart,r4)

		mov.l	@(RenderBuffer_TileArrayBase,r4),r0	! End-of-buffer, aligned 32 bytes, please
		mov.l	@(Render_NativeBufferSize,r8),r5
		mov	#-32,r1
		add	r0,r5
		and	r1,r5

		mova	rdInitRenderBuffer_BucketSizes,r0	! Calculate starting addresses for the
		mov.l	@(Render_BucketSizes,r8),r2		! PP Matrices (located at end of Native Buffer)
		mov	r0,r6

		mov	r15,r7
		add	#5*8,r7
		swap.w	r2,r2
		mov	#5,r1
rdInitRenderBuffer_bucketSize:
		mov	r2,r0
		and	#3,r0
		mov.b	@(r0,r6),r3
		extu.b	r3,r3
		mov.l	r3,@-r7
		mul.l	r3,r9
		sts	macl,r3
		mul.l	r3,r10
		sts	macl,r3
		sub	r3,r5
		tst	r3,r3
		mov	r5,r3
		bf	rdInitRenderBuffer_bucketSizeNonzero
		mov	#1,r3
		rotr	r3
rdInitRenderBuffer_bucketSizeNonzero:
		mov.l	r3,@-r7
		rotl	r2
		rotl	r2
		rotl	r2
		rotl	r2
		dt	r1
		bf	rdInitRenderBuffer_bucketSize

		mov.l	r5,@(RenderBuffer_PPMatrixBase,r4)

		mov	#2,r0					! Some scratch space, in case of overflow later
		shll8	r0
		sub	r0,r5
		mov.l	r5,@(RenderBuffer_PrimAllocEnd,r4)
		mov.l	r5,@(RenderBuffer_PPAllocEnd,r4)

		mov.l	@(RenderBuffer_TileArrayBase,r4),r5	! Initialize Tile Array
		mov	#0xa5,r0
		shll16	r0
		shll8	r0
		or	r0,r5
		mov	#0x80,r3
		shll16	r3
		mov	#0,r2
		shll8	r3
		mov	#0x10,r0
		shll16	r0
		shll8	r0

		mov.l	r0,@(0*4,r5)				! First, a dummy tile with Init bit set
		mov.l	r3,@(1*4,r5)
		mov.l	r3,@(2*4,r5)
		mov.l	r3,@(3*4,r5)
		mov.l	r3,@(4*4,r5)
		mov.l	r3,@(5*4,r5)
		add	#6*4,r5

		mov	#0,r7					! Row-major: good for stripbuffers, bad for AA
rdInitRenderBuffer_tileY:

		mov	#0,r6
rdInitRenderBuffer_tileX:

		mov	r7,r0					! (write tile header)
		shll8	r0
		mov	r6,r1
		shll2	r1
		or	r1,r0
		mov.l	r0,@r5
		add	#4,r5

		mov	r15,r0					! (step down one row / write Tile)
		mov	#5,r3
rdInitRenderBuffer_tileType:
		mov.l	@(0,r0),r1
		mov.l	@(4,r0),r2
		mov.l	r1,@r5
		add	r2,r1
		mov.l	r1,@(0,r0)
		dt	r3
		add	#8,r0
		bf/s	rdInitRenderBuffer_tileType
		add	#4,r5

		mov.l	@(Render_TilesX,r8),r0
		add	#1,r6
		cmp/hs	r0,r6
		bf	rdInitRenderBuffer_tileX

		mov.l	@(Render_TilesY,r8),r0
		add	#1,r7
		cmp/hs	r0,r7
		bf	rdInitRenderBuffer_tileY

		add	#-6*4,r5				! EOL on the last tile
		mov	#0x80,r1
		mov.l	@r5,r0
		shll16	r1
		shll8	r1
		or	r1,r0
		mov.l	r0,@r5

		add	#5*16,r15

		mov.l	@r15+,r8
		mov.l	@r15+,r9
		rts
		mov.l	@r15+,r10

	
				.ALIGN	5
rdInitRenderBuffer_Render:	.LONG	_Render
				.ALIGN	2
rdInitRenderBuffer_BucketSizes:	.BYTE	0*4, 8*4, 16*4, 32*4



!------------------------------------------------------------------------
! void rdSetupVertexBuffer(int buffer, void* base, int length)
!
! Configures one Vertex Buffer
!
! in	r4		int	buffer
!	r5		void*	base
!	r6		int	length
!

		.GLOBL	_rdSetupVertexBuffer

_rdSetupVertexBuffer:
		mov	#5,r2
		mov.l	_rdSetupVertexBuffer_VertexBuffers,r3
		shld	r2,r4
		mov.l	_rdSetupVertexBuffer_P2And,r1
		mov.l	_rdSetupVertexBuffer_P2Or,r2
		add	r4,r3
		and	r1,r5
		mov.l	r6,@(VertexBuffer_Length,r3)
		or	r2,r5
		rts
		mov.l	r5,@(VertexBuffer_Base,r3)

					.ALIGN	3
_rdSetupVertexBuffer_VertexBuffers:	.LONG	_VertexBuffers
_rdSetupVertexBuffer_P2And:		.LONG	0x03ffffff
_rdSetupVertexBuffer_P2Or:		.LONG	0xac000000



!------------------------------------------------------------------------
! void rdBeginScene(bool userEOL)
!
! in	r4		bool	userEOL
!

		.GLOBL	_rdBeginScene
		.ALIGN	5
_rdBeginScene:
		sts.l	pr,@-r15

		mov.l	_rdBeginScene_IntBase,r0
		mov.l	_rdBeginScene_IntClearFlags,r1
		mov.l	r1,@r0
	
		mov.l	_rdEndScene_RenderState,r7
		mov	#TASTATE_ENQUEUEING,r0
		mov.b	r0,@(RenderState_CurrentTAState,r7)

		tst	r4,r4
		mov.l	_rdBeginScene_RenderState,r7
		subc	r0,r0
		add	#1,r0
		mov.b	r0,@(RenderState_Mode,r7)

		mov.l	@(RenderState_CurrentTABuffer,r7),r4 ! Setup TA to reference the appropriate buffer
		bsr	rdSetTABuffer
		nop

		bsr	rdResetVertexBuffers
		nop

		lds.l	@r15+,pr
		rts
		nop

				.ALIGN	2
_rdBeginScene_RenderState:	.LONG	_Render+Render_State
_rdBeginScene_IntBase:		.LONG	0xa05f6900
_rdBeginScene_IntClearFlags:	.LONG	0x00200780



!------------------------------------------------------------------------
! void rdEndScene(void)
!

		.GLOBL	_rdEndScene
		.ALIGN	5
_rdEndScene:
		sts.l	pr,@-r15

		mov.l	_rdEndScene_RenderState,r3		! If UserEOL, do no cleanup at all
		mov.b	@(RenderState_Mode,r3),r0
		tst	#RENDERSTATE_USEREOL,r0
		bf	_rdEndScene_noCleanup
	
		mov.l	@(RenderState_CurrentDirectPath,r3),r1	! If there is a Direct Path, flush + close
		cmp/pz	r1
		bf	_rdEndScene_noActivePath

		bsr	_rdFlushVertexBuffer
		nop

		bsr	_rdFinishDirectPath
		nop
_rdEndScene_noActivePath:

		bsr	rdFlushRemainingListTypes		! Flush all other list types
		nop

_rdEndScene_noCleanup:

		bsr	rdWaitForTADone				! Wait for TA to finish its work
		nop

		mov.l	_rdEndScene_RenderState,r0
		mov.l	@(RenderState_CurrentTABuffer,r0),r4
		bsr	rdAppendBGPlane				! Put BG plane at end of current Prim Allocation
		nop

		bsr	rdCheckVertexBufferOverflow
		nop

		mov.l	_rdEndScene_RenderState,r7
		mov	#TASTATE_DONE,r0
		mov.b	r0,@(RenderState_CurrentTAState,r7)

		lds.l	@r15+,pr
		rts
		nop

				.ALIGN	2
_rdEndScene_RenderState:	.LONG	_Render+Render_State



!------------------------------------------------------------------------
! void rdCheckVertexBufferOverflow(void)
!
! Look through all VertexBuffers and see if any has been overflowed
!

		.ALIGN	5
rdCheckVertexBufferOverflow:

		mov.l	rdCheckVertexBufferOverflow_VertexBuffers,r4
		mov.l	rdCheckVertexBufferOverflow_AreaMask,r3
		mov	#5,r6
rdCheckVertexBufferOverflow_buffer:
		mov.l	@(VertexBuffer_Base,r4),r1
		mov.l	@(VertexBuffer_CurrentWritePtr,r4),r0
		tst	r1,r1
		bt	rdCheckVertexBufferOverflow_skip
		mov.l	@(VertexBuffer_Length,r4),r2
		or	r3,r1
		or	r3,r0
		add	r2,r1
		cmp/hi	r1,r0
		bt	rdCheckVertexBufferOverflow_error

rdCheckVertexBufferOverflow_skip:
		dt	r6
		bf/s	rdCheckVertexBufferOverflow_buffer
		add	#VertexBuffer_SIZEOF,r4

		rts
		nop

rdCheckVertexBufferOverflow_error:
		mova	rdCheckVertexBufferOverflow_errorMsg,r0
		mov	r0,r4
		trapa	#3

							.ALIGN	3
rdCheckVertexBufferOverflow_VertexBuffers:		.LONG	_VertexBuffers
rdCheckVertexBufferOverflow_AreaMask:			.LONG	0xfc000000
							.ALIGN	2
rdCheckVertexBufferOverflow_errorMsg:			.ASCIZ	"rdCheckVertexBufferOverflow: Vertex buffer overflow\r\n"



!------------------------------------------------------------------------
! void rdWaitForTADone(void)
!
! Wait until TA is done enqueueing all the enabled list types
!

		.ALIGN	5
rdWaitForTADone:
		mov.l	rdWaitForTADone_Render,r3
		mov.l	@(Render_EnabledTypes,r3),r1
		mov	#0,r0
		shlr	r1
		bf	rdWaitForTADone_nOpaque
		or	#1,r0
rdWaitForTADone_nOpaque:
		shlr	r1
		bf	rdWaitForTADone_nOpaqueMod
		or	#2,r0
rdWaitForTADone_nOpaqueMod:
		shlr	r1
		bf	rdWaitForTADone_nTranslucent
		or	#4,r0
rdWaitForTADone_nTranslucent:
		shlr	r1
		bf	rdWaitForTADone_nTranslucentMod
		or	#8,r0
rdWaitForTADone_nTranslucentMod:
		shll8	r0
		shlr	r0
		shlr	r1
		bf	rdWaitForTADone_nPunchthrough
		mov	#0x20,r1
		shll16	r1
		or	r1,r0
rdWaitForTADone_nPunchthrough:

		mov.l	rdWaitForTADone_IntBase,r1
rdWaitForTADone_wait:
		mov.l	@r1,r2
		and	r0,r2
		cmp/eq	r0,r2
		bf	rdWaitForTADone_wait
		rts
		nop

				.ALIGN	2
rdWaitForTADone_Render:		.LONG	_Render
rdWaitForTADone_IntBase:	.LONG	0xa05f6900


	
!------------------------------------------------------------------------
! void rdSetTABuffer(RenderBuffer* buffer)
!
! in	r4		RenderBuffer* buffer
!
		.ALIGN	5	
rdSetTABuffer:
		mov.l	rdSetTABuffer_Render,r0
		mov	r0,r6
		mov.l	rdSetTABuffer_IntBase,r0
		mov	#-1,r1
		mov.l	r1,@r0

		mov.l	rdSetTABuffer_PVR2Regs,r7

		mov	#1,r0					! Reset TA
		mov.l	r0,@(COREDISABLE*4,r7)
		mov	#0,r0					! Un-reset TA
		mov.l	r0,@(COREDISABLE*4,r7)
	
		mov.l	@(RenderBuffer_PPMatrixBase,r4),r1	! (TA) Set base address for the root PP-block matrices
		mov	#PPMATRIXBASE,r0
		shll2	r0
		mov.l	r1,@(r0,r7)

		mov.l	@(RenderBuffer_PrimAllocStart,r4),r1	! (TA) Set allocation range for PP-blocks
		mov	#PRIMALLOCSTART,r0
		shll2	r0
		mov.l	r1,@(r0,r7)
		mov.l	@(RenderBuffer_PrimAllocEnd,r4),r1
		mov	#PRIMALLOCEND,r0
		shll2	r0
		mov.l	r1,@(r0,r7)

		mov.l	@(RenderBuffer_PPAllocStart,r4),r1	! (TA) Set allocation range for primitives
		mov	#PPALLOCSTART,r0
		shll2	r0
		mov.l	r1,@(r0,r7)
		mov.l	@(RenderBuffer_PPAllocEnd,r4),r1
		mov	#PPALLOCEND,r0
		shll2	r0
		mov.l	r1,@(r0,r7)

		mov.l	@(Render_TilesX,r6),r1			! (TA) Tile array dimensions
		mov.l	@(Render_TilesY,r6),r2
		add	#-1,r1
		add	#-1,r2
		shll16	r2
		or	r2,r1
		mov	#TASIZE,r0
		shll2	r0
		mov.l	r1,@(r0,r7)

		mov.l	@(Render_BucketSizes,r6),r1		! (TA) PP-block sizes
		mov	#0x10,r2
		shll16	r2
		or	r2,r1
		mov	#PPBLOCKSIZE,r0
		shll2	r0
		mov.l	r1,@(r0,r7)

		mov	#TASTART,r0				! (TA) Roll it - time for new vertex registration
		shll2	r0
		mov.l	r0,@(r0,r7)
	
		rts
		nop
	
			.ALIGN	3
rdSetTABuffer_Render:	.LONG	_Render
rdSetTABuffer_IntBase:	.LONG	0xa05f6900
rdSetTABuffer_PVR2Regs:	.LONG	PVR2REGS



!------------------------------------------------------------------------
! void rdAppendBGPlane(RenderBuffer* buffer)
!
! in	r4		RenderBuffer*	buffer
!

		.ALIGN	5	
rdAppendBGPlane:
		mov.l	rdAppendBGPlane_PVR2Regs,r7

		mov	#PRIMALLOCPOS,r0		! (TA) Current Prim Allocation position
		shll2	r0
		mov.l	@(r0,r7),r6

		mov.l	r6,@(RenderBuffer_BGPlaneAddr,r4)

		mov	r6,r2				! Stuff BG plane into graphics memory,
		mov	#0xa5,r1			! just after current Prim Allocation
		shll16	r1
		shll8	r1
		or	r2,r1
		mov.l	rdAppendBGPlane_BGPlane,r0
		mov	#15,r3
CopyLong:
		mov.l	@r0+,r2
		mov.l	r2,@r1
		dt	r3
		bf/s	CopyLong
		add	#4,r1

		rts
		nop

				.ALIGN	2
rdAppendBGPlane_BGPlane:	.LONG	BGPlane
rdAppendBGPlane_PVR2Regs:	.LONG	PVR2REGS



!------------------------------------------------------------------------
! void rdSet3DBuffer(RenderBuffer* buffer)
!
! in	r4	RenderBuffer* buffer
!

		.ALIGN	5	
rdSet3DBuffer:
		mov.l	rdSet3DBuffer_Render,r0
		mov.l	rdSet3DBuffer_PVR2Regs,r7
		mov	r0,r6

		mov	#-12,r3
		mov	#-1,r1					! (3D) BG plane location
		mov.l	@(RenderBuffer_BGPlaneAddr,r4),r2
		shld	r3,r1
		mov.l	@(RenderBuffer_PrimAllocStart,r4),r3
		neg	r1,r1
		and	r1,r3
		sub	r3,r2
		shll	r2
		shll8	r2
		shlr8	r2
		mov	#0x01,r1
		shll16	r1
		shll8	r1
		or	r2,r1
		mov	#BGPLANE,r0
		shll2	r0
		mov.l	r1,@(r0,r7)

		mov	#PRIMALLOCBASE,r0			! (3D) MB-number of primitive allocation area
		shll2	r0
		mov.l	r3,@(r0,r7)

		mov.l	@(RenderBuffer_TileArrayBase,r4),r1	! (3D) Tile Array
		mov	#TABASE,r0			
		shll2	r0
		mov.l	r1,@(r0,r7)

		mov.l	@(RenderBuffer_ScreenBase,r4),r1	! (3D) Output base address
		mov	#RENDERBASE,r0
		shll2	r0
		mov.l	r1,@(r0,r7)

		mov.l	@(Render_TargetPitch,r6),r1		! (3D) Output pitch
		shlr2	r1
		shlr	r1
		mov	#RENDERPITCH,r0
		shll2	r0
		mov.l	r1,@(r0,r7)

		mov.l	@(Render_TargetFormat,r6),r1		! (3D) Output format
		mov	#8,r2
		or	r2,r1					! enable 32->16bit dither
		mov	#RENDERFORMAT,r0
		shll2	r0
		mov.l	r1,@(r0,r7)

		mov.l	@(Render_TargetWidth,r6),r1		! (3D) Output clip window X
		add	#-1,r1
		shll16	r1
		mov	#RENDERWINDOWX,r0
		shll2	r0
		mov.l	r1,@(r0,r7)

		mov.l	@(Render_TargetHeight,r6),r1		! (3D) Output clip window Y
		add	#-1,r1
		shll16	r1
		mov	#RENDERWINDOWY,r0
		shll2	r0
		mov.l	r1,@(r0,r7)

		mov	#RENDERSTART,r0				! (3D) kick it. commence rendering. all bets are off.
		shll2	r0
		rts
		mov.l	r0,@(r0,r7)

				.ALIGN	2
rdSet3DBuffer_Render:		.LONG	_Render
rdSet3DBuffer_PVR2Regs:		.LONG	PVR2REGS



!------------------------------------------------------------------------
! void rdFlushRemainingListTypes(void)
!
! All list types which haven't been flushed/EOLed yet: do that
!

		.ALIGN	5
rdFlushRemainingListTypes:
		mov.l	r9,@-r15
		mov.l	r8,@-r15
		sts.l	pr,@-r15
		mov.l	rdFlushRemainingListTypes_Render,r6
		mov	r6,r7
		add	#Render_State,r7
		mov.l	@(Render_EnabledTypes,r6),r8
		mov.l	@(RenderState_FinishedPaths,r7),r1
		not	r1,r1
		and	r1,r8

		mov	#0,r9
rdFlushRemainingListTypes_type:
		shlr	r8
		bf	rdFlushRemainingListTypes_skipType

		bsr	_rdSetupDirectPath
		mov	r9,r4

		bsr	_rdFinishDirectPath
		nop

rdFlushRemainingListTypes_skipType:
		add	#1,r9
		mov	#5,r0
		cmp/hs	r0,r9
		bf	rdFlushRemainingListTypes_type

		lds.l	@r15+,pr
		mov.l	@r15+,r8
		rts
		mov.l	@r15+,r9

					.ALIGN	2
rdFlushRemainingListTypes_Render:	.LONG	_Render


	
!------------------------------------------------------------------------
! void rdResetVertexBuffers(void)
!
! Reset all Vertex Buffers to empty state, as well as the Pathing
!

		.ALIGN	5
rdResetVertexBuffers:
		mov.l	rdResetVertexBuffers_VertexBuffers,r4
		mov.l	rdResetVertexBuffers_SQAnd,r5
		mov.l	rdResetVertexBuffers_SQOr,r6
		mov	#5,r3
rdResetVertexBuffers_resetVertexBuffer:
		mov.l	@(VertexBuffer_Base,r4),r2
		dt	r3

		mov	r2,r1
		shlr16	r2
		shlr8	r2
		and	r5,r1
		mov.l	r2,@(VertexBuffer_CurrentWriteArea,r4)
		or	r6,r1
		mov.l	r1,@(VertexBuffer_CurrentWritePtr,r4)
		mov.l	r1,@(VertexBuffer_CurrentFlushPos,r4)
		bf/s	rdResetVertexBuffers_resetVertexBuffer
		add	#VertexBuffer_SIZEOF,r4

		mov.l	rdResetVertexBuffers_RenderState,r0
		mov	#-1,r1
		mov.l	r1,@(RenderState_CurrentDirectPath,r0)
		mov	#0,r1
		rts
		mov.l	r1,@(RenderState_FinishedPaths,r0)

					.ALIGN	4
rdResetVertexBuffers_VertexBuffers:	.LONG	_VertexBuffers
rdResetVertexBuffers_RenderState:	.LONG	_Render+Render_State
rdResetVertexBuffers_SQAnd:		.LONG	0x03ffffff
rdResetVertexBuffers_SQOr:		.LONG	0xe0000000



!------------------------------------------------------------------------
! void rdSetupDirectPath(int type)
!
! Make the given list type feed directly into TA
!
! in	r4	int	type
!
		.GLOBL	_rdSetupDirectPath
		.ALIGN	5
_rdSetupDirectPath:
		mov.l	_rdSetupDirectPath_Render,r6
		mov	r6,r7
		add	#Render_State,r7

		mov.l	@(RenderState_CurrentDirectPath,r7),r0		! Sanity checks
		cmp/pz	r0
		bt	_rdSetupDirectPath_error1

		mov	#1,r1
		shld	r4,r1
		mov.l	@(RenderState_FinishedPaths,r7),r2
		tst	r1,r2
		bf	_rdSetupDirectPath_error2

		mov.l	@(Render_EnabledTypes,r6),r2
		tst	r1,r2
		bt	_rdSetupDirectPath_error3

		mov.l	r4,@(RenderState_CurrentDirectPath,r7)		! Select list type as direct path

		mov	#5,r3
		mov.l	_rdSetupDirectPath_DirectPathSavedWritePtr,r1	! Backup WritePtr; set WritePtr
		shld	r3,r4						! directly to TA
		mov.l	_rdSetupDirectPath_VertexBuffers,r2
		mov	#0x10,r0
		add	r4,r2
		mov.l	@(VertexBuffer_CurrentWritePtr,r2),r3
		mov.l	r0,@(VertexBuffer_CurrentWriteArea,r2)
		rts
		mov.l	r3,@r1

_rdSetupDirectPath_error1:
		mova	_rdSetupDirectPath_ErrorMsg1,r0
		mov	r0,r4
		trapa	#3
_rdSetupDirectPath_error2:
		mova	_rdSetupDirectPath_ErrorMsg2,r0
		mov	r0,r4
		trapa	#3
_rdSetupDirectPath_error3:
		mova	_rdSetupDirectPath_ErrorMsg3,r0
		mov	r0,r4
		trapa	#3

						.ALIGN	4
_rdSetupDirectPath_Render:			.LONG	_Render
_rdSetupDirectPath_DirectPathSavedWritePtr:	.LONG	DirectPathSavedWritePtr
_rdSetupDirectPath_VertexBuffers:		.LONG	_VertexBuffers
						.ALIGN	2
_rdSetupDirectPath_ErrorMsg1:			.ASCIZ	"rdSetupDirectPath: Another list type is direct path already\r\n"
						.ALIGN	2
_rdSetupDirectPath_ErrorMsg2:			.ASCIZ	"rdSetupDirectPath: List type has already been finished\r\n"
						.ALIGN	2
_rdSetupDirectPath_ErrorMsg3:			.ASCIZ	"rdSetupDirectPath: List type is not enabled\r\n"


		
!------------------------------------------------------------------------
! void rdFinishDirectPath(void)
!
! Finish the list type that is currently direct path'ed; flush its vertex buffer
!

		.GLOBL	_rdFinishDirectPath
		.ALIGN	5
_rdFinishDirectPath:
		mov.l	_rdFinishDirectPath_RenderState,r3

		mov.l	@(RenderState_CurrentDirectPath,r3),r0
		cmp/pz	r0
		bf	_rdFinishDirectPath_error

		sts.l	pr,@-r15

		bsr	_rdFlushVertexBuffer				! Flush VertexBuffer for the path
		nop

		mov.l	_rdFinishDirectPath_RenderState,r3		! Restore WritePtr for the path
		mov.l	_rdFinishDirectPath_DirectPathSavedWritePtr,r1
		mov	#0xac,r6
		mov.l	@(RenderState_CurrentDirectPath,r3),r0
		mov	#3,r5
		mov	r0,r4
		shll2	r0
		mov.l	@r1,r1
		shld	r5,r0
		mov.l	_rdFinishDirectPath_VertexBuffers,r2
		add	r0,r2
		mov.l	r1,@(VertexBuffer_CurrentWritePtr,r2)
		mov.l	r6,@(VertexBuffer_CurrentWriteArea,r2)

		bsr	rdSendTAEOL					! Send striphead + EOL to TA (in case list is empty)
		nop

		lds.l	@r15+,pr
	
		mov.l	_rdFinishDirectPath_RenderState,r3		! Mark list type as finished,
		mov.l	@(RenderState_CurrentDirectPath,r3),r0		! and set no current path as active
		mov.l	@(RenderState_FinishedPaths,r3),r2
		mov	#1,r1
		shld	r0,r1
		mov	#-1,r0
		or	r1,r2
		mov.l	r2,@(RenderState_FinishedPaths,r3)
		rts
		mov.l	r0,@(RenderState_CurrentDirectPath,r3)

_rdFinishDirectPath_error:
		mova	_rdFinishDirectPath_ErrorMsg,r0
		mov	r0,r4
		trapa	#3

						.ALIGN	4
_rdFinishDirectPath_RenderState:		.LONG	_Render+Render_State
_rdFinishDirectPath_DirectPathSavedWritePtr:	.LONG	DirectPathSavedWritePtr
_rdFinishDirectPath_VertexBuffers:		.LONG	_VertexBuffers
						.ALIGN	2
_rdFinishDirectPath_ErrorMsg:			.ASCIZ	"rdFinishDirectPath: No list type is direct path right now\r\n"



!------------------------------------------------------------------------
! void rdFlushVertexBuffer(void)
!
! Flush the vertex buffer of the Direct Path
!

		.GLOBL	_rdFlushVertexBuffer
		.ALIGN	5
_rdFlushVertexBuffer:
		mov.l	_rdFlushVertexBuffer_RenderState,r3

		mov.l	@(RenderState_CurrentDirectPath,r3),r0
		cmp/pz	r0
		bf	_rdFlushVertexBuffer_error

		mov	#5,r2
		mov.l	_rdFlushVertexBuffer_VertexBuffers,r1		! Check if VertexBuffer is non-empty
		shld	r2,r0
		mov.l	_rdFlushVertexBuffer_DirectPathSavedWritePtr,r3
		add	r0,r1
		mov.l	@(VertexBuffer_Base,r1),r0
		mov.l	@(VertexBuffer_CurrentFlushPos,r1),r4
		mov.l	@r3,r5
		tst	r0,r0						! Exists at all?
		bt	_rdFlushVertexBuffer_zeroBytes
		cmp/eq	r4,r5
		bt	_rdFlushVertexBuffer_zeroBytes			! Non-empty?

		mov.l	r5,@(VertexBuffer_CurrentFlushPos,r1)		! Advance FlushPtr
		sub	r4,r5

		sts.l	pr,@-r15
		mov.l	_rdFlushVertexBuffer_P2And,r0
		mov.l	_rdFlushVertexBuffer_P2Or,r1
		and	r0,r4
		bsr	rdSendTANonCachedStream				! Flush out data
		or	r1,r4
		lds.l	@r15+,pr

_rdFlushVertexBuffer_zeroBytes:

		rts
		nop

_rdFlushVertexBuffer_error:
		mova	_rdFlushVertexBuffer_ErrorMsg,r0
		mov	r0,r4
		trapa	#3

						.ALIGN	4
_rdFlushVertexBuffer_VertexBuffers:		.LONG	_VertexBuffers
_rdFlushVertexBuffer_DirectPathSavedWritePtr:	.LONG	DirectPathSavedWritePtr
_rdFlushVertexBuffer_RenderState:		.LONG	_Render+Render_State
_rdFlushVertexBuffer_P2And:			.LONG	0x03ffffff
_rdFlushVertexBuffer_P2Or:			.LONG	0xac000000
						.ALIGN	2
_rdFlushVertexBuffer_ErrorMsg:			.ASCIZ	"rdFlushVertexBuffer: No direct path active\r\n"



!------------------------------------------------------------------------
! void rdSendTAEOL(int type)
!
! Send dummy striphead + EOL to TA
!

		.ALIGN	5
rdSendTAEOL:
		mov	#0xff,r0
		shll16	r0
		shll8	r0
		or	#0x38,r0
		mov	#0x10,r1
		mov.l	r1,@(0,r0)
		mov.l	r1,@(4,r0)

		mov	#0xe0,r0
		shll16	r0
		mov	#0,r1
		shll8	r0

		add	#0x80,r4
		shll16	r4
		mov	#0,r1
		shll8	r4
		mov.l	r4,@r0
		pref	@r0
		add	#32,r0
		mov.l	r1,@r0
		rts
		pref	@r0



!------------------------------------------------------------------------
! void rdSendTANonCachedStream(void* start, int length)
!
! Send stream of data to TA (cacheline granularity)
!
! in	r4	void*	start of data (in P2 area)
!	r5	int	number of bytes
!

		.ALIGN	5
rdSendTANonCachedStream:
		tst	r5,r5
		bt	_rdSendTANonCachedStream_zeroBytes

		mov	#0x90,r0
		shll16	r0
		shll8	r0
		add	#-1,r0
		and	r0,r4

		pref	@r4
		mov	#0xff,r0
		shll16	r0
		shll8	r0
		or	#0x38,r0
		mov	#0x10,r1
		mov.l	r1,@(0,r0)
		mov.l	r1,@(4,r0)

		mov	#0xe0,r0
		shll16	r0
		shll8	r0

		fschg
		shlr2	r5
		shlr2	r5
		shlr	r5

_rdSendTANonCachedStream_transferLine:
		fmov.d	@r4+,dr0
		fmov.d	@r4+,dr2
		fmov.d	@r4+,dr4
		fmov.d	@r4+,dr6
		pref	@r4
		add	#-4,r4
		fmov.d	dr0,@r0
		add	#8,r0
		fmov.d	dr2,@r0
		add	#8,r0
		fmov.d	dr4,@r0
		add	#8,r0
		fmov.d	dr6,@r0
		dt	r5
		pref	@r0
		add	#8,r0
		ocbi	@r4
		bf/s	_rdSendTANonCachedStream_transferLine
		add	#4,r4

		ocbi	@r4
		fschg

_rdSendTANonCachedStream_zeroBytes:
		rts
		nop



!------------------------------------------------------------------------
! void rdDrawPrimitives(int type, void* start, int length)
!
! Enqueue stream of primitives (cacheline granularity)
!
! in	r4	int	list type
!	r5	void*	start of data (in P1 area)
!	r6	int	number of bytes
!

		.GLOBL	_rdDrawPrimitives
		.ALIGN	5
_rdDrawPrimitives:
		mov.l	_rdDrawPrimitives_VertexBuffers,r0
		mov	#5,r1
		tst	r6,r6
		shld	r1,r4
		bt	_rdDrawPrimitives_zeroBytes

		add	r0,r4
		mov.l	@(VertexBuffer_CurrentWritePtr,r4),r1
		mov.l	_rdDrawPrimitives_QACR,r3
		mov	r1,r2
		mov.l	@(VertexBuffer_CurrentWriteArea,r4),r0
		add	r6,r2
		mov.l	r2,@(VertexBuffer_CurrentWritePtr,r4)

		mov.l	r0,@(0,r3)
		mov	#-5,r2
		mov.l	r0,@(4,r3)
		shld	r2,r6

		fschg
		add	#-8,r1

_rdDrawPrimitives_transferLine:
		fmov.d	@r5+,dr0
		add	#8,r1
		fmov.d	@r5+,dr2
		fmov.d	@r5+,dr4
		fmov.d	@r5+,dr6
		pref	@r5
		dt	r6
		fmov.d	dr0,@r1
		add	#8,r1
		fmov.d	dr2,@r1
		add	#8,r1
		fmov.d	dr4,@r1
		add	#8,r1
		fmov.d	dr6,@r1
		bf/s	_rdDrawPrimitives_transferLine
		pref	@r1

		fschg

_rdDrawPrimitives_zeroBytes:
		rts
		nop

					.ALIGN	4
_rdDrawPrimitives_VertexBuffers:	.LONG	_VertexBuffers
_rdDrawPrimitives_QACR:			.LONG	0xff000038



!------------------------------------------------------------------------
! void rdIRQ9Handler(int offset, int evt)

rdIRQ9Handler:
		sts.l	pr,@-r15


		mov.l	rdIRQ9Handler_rdIRQ9EventTable,r4
		mov.l	rdIRQ9Handler_IntBase,r5
		bsr	rdIRQ9Handler_ScanForIRQs
		nop

		mov.l	rdIRQ9Handler_IntBase,r5
		mov.l	rdIRQ9Handler_rdIRQ9EventTable2,r4
		bsr	rdIRQ9Handler_ScanForIRQs
		add	#8,r5

		lds.l	@r15+,pr
		rts
		nop

					.ALIGN	4
rdIRQ9Handler_rdIRQ9EventTable:		.LONG	rdIRQ9EventTable
rdIRQ9Handler_rdIRQ9EventTable2:	.LONG	rdIRQ9EventTable+32*4
rdIRQ9Handler_IntBase:			.LONG	0xa05f6900


!	r4	func[]	handlers
!	r5	uint32*	irqreg

		.ALIGN	5
rdIRQ9Handler_ScanForIRQs:
		mov.l	@(0x30,r5),r7

		mov.l	@r5,r0
		and	r7,r0
		tst	r0,r0
		bt	rdIRQ9Handler_noInts

		mov	#-1,r1
		shll16	r1
		mov	#0,r6
		tst	r1,r0
		bt	rdIRQ9Handler_ScanForIRQs_low16
		shlr16	r0
		add	#16,r6
rdIRQ9Handler_ScanForIRQs_low16:
		shlr8	r1
		tst	r1,r0
		bt	rdIRQ9Handler_ScanForIRQs_low8
		shlr8	r0
		add	#8,r6
rdIRQ9Handler_ScanForIRQs_low8:
		tst	#0xf0,r0
		bt	rdIRQ9Handler_ScanForIRQs_low4
		shlr2	r0
		shlr2	r0
		add	#4,r6
rdIRQ9Handler_ScanForIRQs_low4:
		tst	#0xc,r0
		bt	rdIRQ9Handler_ScanForIRQs_low2
		shlr2	r0
		add	#2,r6
rdIRQ9Handler_ScanForIRQs_low2:
		tst	#2,r0
		movt	r1
		add	#1,r6
		sub	r1,r6

		mov	r6,r0
		shll2	r0
		mov.l	@(r0,r4),r0
		jmp	@r0		! Called with (functable, irqreg, intbit)
		nop

rdIRQ9Handler_noInts:

		rts
		nop


		.ALIGN	5
rdIRQ9Event_Ignore:
		mov	#1,r0
		shld	r6,r0
		rts
		mov.l	r0,@r5


		.ALIGN	5
rdIRQ9Event_Default:
		sts.l	pr,@-r15

		mov	#1,r0
		shld	r6,r0
		mov.l	r0,@r5
	
		mov	r5,r0
		and	#8,r0
		shll2	r0
		add	r0,r6

		mov.l	r6,@-r15

		mov.l	rdIRQ9Event_Default_SerialWriteString,r1
		mova	rdIRQ9Event_Default_String1,r0
		jsr	@r1
		mov	r0,r4
		
		mov.l	rdIRQ9Event_Default_SerialWriteHexByte,r1
		jsr	@r1
		mov.l	@r15+,r4

		mov.l	rdIRQ9Event_Default_SerialWriteString,r1
		mova	rdIRQ9Event_Default_String2,r0
		jsr	@r1
		mov	r0,r4

		lds.l	@r15+,pr
		rts
		nop

						.ALIGN	2
rdIRQ9Event_Default_SerialWriteString:		.LONG	_SerialWriteString
rdIRQ9Event_Default_SerialWriteHexByte:		.LONG	_SerialWriteHexByte
						.ALIGN	2
rdIRQ9Event_Default_String1:			.ASCIZ	"***IRQ9 bit 0x"
						.ALIGN	2
rdIRQ9Event_Default_String2:			.ASCIZ	" triggered\r\n"


		.ALIGN	5
rdIRQ9Event_ScanlineInt1:			! Scanline interrupt
		mov	#1,r0
		shld	r6,r0
		mov.l	r0,@r5

		sts.l	pr,@-r15

		mov.l	rdIRQ9Event_ScanlineInt1_ScanlineInt1UserCallback,r0
		mov.l	@r0,r0
		tst	r0,r0
		bt	rdIRQ9Event_ScanlineInt1_userCallbackDisabled
!		jsr	@r0
		nop
rdIRQ9Event_ScanlineInt1_userCallbackDisabled:

		lds.l	@r15+,pr
		rts
		nop

		.ALIGN	2
rdIRQ9Event_ScanlineInt1_ScanlineInt1UserCallback: .LONG ScanlineInt1UserCallback


		.ALIGN	5
rdIRQ9Event_ScanlineInt2:			! Scanline interrupt
		mov	#1,r0
		shld	r6,r0
		mov.l	r0,@r5

		sts.l	pr,@-r15

		mov.l	rdIRQ9Event_ScanlineInt2_ScanlineInt2UserCallback,r0
		mov.l	@r0,r0
		tst	r0,r0
		bt	rdIRQ9Event_ScanlineInt2_userCallbackDisabled
!		jsr	@r0
		nop
rdIRQ9Event_ScanlineInt2_userCallbackDisabled:

		lds.l	@r15+,pr
		rts
		nop

		.ALIGN	2
rdIRQ9Event_ScanlineInt2_ScanlineInt2UserCallback: .LONG ScanlineInt2UserCallback


		.ALIGN	5
rdIRQ9Event_VSync:				! Vertical retrace interrupt
		mov	#1,r0
		mov.l	rdIRQ9Event_VSync_FrameCounter,r1
		shld	r6,r0
		mov.l	r0,@r5
		mov.l	@r1,r0
		add	#1,r0
		mov.l	r0,@r1

		sts.l	pr,@-r15

		bsr	rdRenderVSyncHandler
		nop

		mov.l	rdIRQ9Event_VSync_VSyncUserCallback,r0
		mov.l	@r0,r0
		tst	r0,r0
		bt	rdIRQ9Event_VSync_userCallbackDisabled
		jsr	@r0
		nop
rdIRQ9Event_VSync_userCallbackDisabled:

		lds.l	@r15+,pr

		rts
		nop
	
					.ALIGN	3
rdIRQ9Event_VSync_VSyncUserCallback:	.LONG	VSyncUserCallback
rdIRQ9Event_VSync_FrameCounter:		.LONG	_FrameCounter


	
		.ALIGN	5
rdIRQ9Event_RenderDone:				! 3D rendering unit done interrupt
		mov	#1,r0
		shld	r6,r0
		mov.l	r0,@r5

		sts.l	pr,@-r15

		bsr	rdRenderDoneHandler
		nop

		mov.l	rdIRQ9Event_RenderDone_RenderDoneUserCallback,r0
		mov.l	@r0,r0
		tst	r0,r0
		bt	rdIRQ9Event_RenderDone_userCallbackDisabled
		jsr	@r0
		nop
rdIRQ9Event_RenderDone_userCallbackDisabled:

		lds.l	@r15+,pr

		rts
		nop
	
						.ALIGN	2
rdIRQ9Event_RenderDone_RenderDoneUserCallback:	.LONG	RenderDoneUserCallback


		.ALIGN	5
rdIRQ9Event_FatalError:				! TA/3D rendering unit fatal error
		mov	#1,r0
		shld	r6,r0
		mov.l	r0,@r5

		sts.l	pr,@-r15
		shll2	r6
		mova	rdIRQ9Event_FatalError_Strings,r0
		mov.l	rdIRQ9Event_FatalError_SerialWriteString,r1
		jsr	@r1
		mov.l	@(r0,r6),r4
		lds.l	@r15+,pr
		rts
		nop

						.ALIGN	5
rdIRQ9Event_FatalError_SerialWriteString:	.LONG	_SerialWriteString
rdIRQ9Event_FatalError_Strings:			.LONG	rdIRQ9Event_FatalError_Error0
						.LONG	0
						.LONG	rdIRQ9Event_FatalError_Error2
						.LONG	rdIRQ9Event_FatalError_Error3

				.ALIGN	2
rdIRQ9Event_FatalError_Error0:	.ASCIZ	"***IRQ9 bit 0x00: 3D rendering unit is really unhappy!***\r\n"
rdIRQ9Event_FatalError_Error2:	.ASCIZ	"***IRQ9 bit 0x02: Prim buffer overflow!***\r\n"
rdIRQ9Event_FatalError_Error3:	.ASCIZ	"***IRQ9 bit 0x03: PP buffer overflow!***\r\n"

		.ALIGN	5
rdIRQ9Event_Test:
		sts.l	pr,@-r15

		mov.l	rdIRQ9Event_Test_Beampos,r1
		mov.l	rdIRQ9Event_Test_SerialWriteHexLong,r0
		jsr	@r0
		mov.l	@r1,r4

		lds.l	@r15+,pr
		rts
		nop

						.ALIGN	2
rdIRQ9Event_Test_SerialWriteHexLong:		.LONG	_SerialWriteHexLong
rdIRQ9Event_Test_Beampos:			.LONG	PVR2REGS+BEAMPOS*4

			.ALIGN	5
rdIRQ9EventTable:	.LONG	rdIRQ9Event_Default
			.LONG	rdIRQ9Event_Default
			.LONG	rdIRQ9Event_RenderDone
			.LONG	rdIRQ9Event_ScanlineInt1
			.LONG	rdIRQ9Event_ScanlineInt2
			.LONG	rdIRQ9Event_VSync
			.REPT	32-6
			.LONG	rdIRQ9Event_Default
			.ENDR
			.LONG	rdIRQ9Event_FatalError
			.LONG	rdIRQ9Event_Ignore
			.LONG	rdIRQ9Event_FatalError
			.LONG	rdIRQ9Event_FatalError
			.REPT	32-4
			.LONG	rdIRQ9Event_Default
			.ENDR


!------------------------------------------------------------------------
! void rdSetRenderDoneUserCallback(void(*)() function)
!
! in	r4	void(*)()	function to call (0 == disable)
!

		.GLOBL	_rdSetRenderDoneUserCallback
		.ALIGN	5
_rdSetRenderDoneUserCallback:
		mov.l	_rdSetRenderDoneUserCallback_RenderDoneUserCallback,r0
		rts
		mov.l	r4,@r0

							.ALIGN	2
_rdSetRenderDoneUserCallback_RenderDoneUserCallback:	.LONG	RenderDoneUserCallback



!------------------------------------------------------------------------
! void rdSetVSyncUserCallback(void(*)() function)
!
! in	r4	void(*)()	function to call (0 == disable)
!

		.GLOBL	_rdSetVSyncUserCallback
		.ALIGN	5
_rdSetVSyncUserCallback:
		mov.l	_rdSetVSyncUserCallback_VSyncUserCallback,r0
		rts
		mov.l	r4,@r0

						.ALIGN	2
_rdSetVSyncUserCallback_VSyncUserCallback:	.LONG	VSyncUserCallback



!------------------------------------------------------------------------
! void rdSetScanlineInt1UserCallback(void(*)() function)
!
! in	r4	void(*)()	function to call (0 == disable)
!

		.GLOBL	_rdSetScanlineInt1UserCallback
		.ALIGN	5
_rdSetScanlineInt1UserCallback:
		mov.l	_rdSetScanlineInt1UserCallback_ScanlineInt1UserCallback,r0
		rts
		mov.l	r4,@r0

								.ALIGN	2
_rdSetScanlineInt1UserCallback_ScanlineInt1UserCallback:	.LONG	ScanlineInt1UserCallback



!------------------------------------------------------------------------
! void rdSetScanlineInt2UserCallback(void(*)() function)
!
! in	r4	void(*)()	function to call (0 == disable)
!

		.GLOBL	_rdSetScanlineInt2UserCallback
		.ALIGN	5
_rdSetScanlineInt2UserCallback:
		mov.l	_rdSetScanlineInt2UserCallback_ScanlineInt2UserCallback,r0
		rts
		mov.l	r4,@r0

								.ALIGN	2
_rdSetScanlineInt2UserCallback_ScanlineInt2UserCallback:	.LONG	ScanlineInt2UserCallback



!------------------------------------------------------------------------
! void rdWaitTAIdle(void)
!
! Idle until TA reaches 'Idle' state (either directly, or after N vsyncs)
!

		.GLOBL	_rdWaitTAIdle
		.ALIGN	5
_rdWaitTAIdle:
		mov.l	_rdWaitTAIdle_RenderState,r1
		mov.b	@(RenderState_CurrentTAState,r1),r0
		cmp/eq	#TASTATE_IDLE,r0
		bf	_rdWaitTAIdle
		rts
		nop

				.ALIGN	2
_rdWaitTAIdle_RenderState:	.LONG	_Render+Render_State

	
		
!------------------------------------------------------------------------
! void rdWaitRenderIdle(void)
!
! Idle until 3D rendering unit reaches 'Idle' state
!

		.GLOBL	_rdWaitRenderIdle
		.ALIGN	5
_rdWaitRenderIdle:
		mov.l	_rdWaitRenderIdle_RenderState,r1
		mov.b	@(RenderState_CurrentRenderState,r1),r0
		cmp/eq	#RENDERSTATE_IDLE,r0
		bf	_rdWaitRenderIdle
		rts
		nop

				.ALIGN	2
_rdWaitRenderIdle_RenderState:	.LONG	_Render+Render_State



!------------------------------------------------------------------------
! void rdFlip(void)
!
! If any rendering has commenced, flip buffers; wait for VSync;
! then wait for TA Idle
!

		.GLOBL	_rdFlip
		.ALIGN	5
_rdFlip:
		mov.l	_rdFlip_FrameCounter,r1
		mov.l	@r1,r2
_rdFlip_waitVSync:
		mov.l	@r1,r0
		cmp/eq	r0,r2
		bt	_rdFlip_waitVSync

		mov.l	_rdFlip_RenderState,r1
_rdFlip_waitTAIdle:
		mov.b	@(RenderState_CurrentTAState,r1),r0
		cmp/eq	#TASTATE_IDLE,r0
		bf	_rdFlip_waitTAIdle

		rts
		nop

			.ALIGN	2
_rdFlip_FrameCounter:	.LONG	_FrameCounter
_rdFlip_RenderState:	.LONG	_Render+Render_State


		.ALIGN	5
rdRenderVSyncHandler:

		sts.l	pr,@-r15

		mov.l	rdRenderVSyncHandler_RenderState,r7
		mov.b	@(RenderState_CurrentRenderState,r7),r0
		cmp/eq	#RENDERSTATE_DONE,r0
		bf	rdRenderVSyncHandler_nPrevRenderDone

		mov	#RENDERSTATE_IDLE,r0				! Swap 3d & 2d buffers
		mov.b	r0,@(RenderState_CurrentRenderState,r7)
		mov	#1,r3
		mov.b	@(RenderState_Current3DBufferIdx,r7),r0
		xor	#1,r0
		mov.b	r0,@(RenderState_Current3DBufferIdx,r7)
		mov	r0,r1
		mov.b	@(RenderState_Current2DBufferIdx,r7),r0
		shll2	r1
		xor	#1,r0
		mov.b	r0,@(RenderState_Current2DBufferIdx,r7)
		mov	r0,r2
		mov.l	rdRenderVSyncHandler_RenderBuffers,r0
		shll2	r2
		mov.l	@(r0,r1),r1
		mov.l	@(r0,r2),r2
		mov.l	r1,@(RenderState_Current3DBuffer,r7)
		mov.l	r2,@(RenderState_Current2DBuffer,r7)

		mov.l	rdRenderVSyncHandler_Render,r0
		mov.l	rdRenderVSyncHandler_scSetBuffer,r2
		mov.l	@(RenderState_Current2DBuffer,r7),r1
		mov.l	@(Render_TargetPitch,r0),r5
		jsr	@r2
		mov.l	@(RenderBuffer_ScreenBase,r1),r4
	
rdRenderVSyncHandler_nPrevRenderDone:

		mov.l	rdRenderVSyncHandler_RenderState,r7
		mov.b	@(RenderState_CurrentTAState,r7),r0
		cmp/eq	#TASTATE_DONE,r0
		bf	rdRenderVSyncHandler_nStartNewRender
		mov.b	@(RenderState_CurrentRenderState,r7),r0
		cmp/eq	#RENDERSTATE_IDLE,r0
		bf	rdRenderVSyncHandler_nStartNewRender

		mov	#TASTATE_IDLE,r0
		mov.b	r0,@(RenderState_CurrentTAState,r7)
		mov	#RENDERSTATE_RENDERING,r0
		mov.b	r0,@(RenderState_CurrentRenderState,r7)

		mov.b	@(RenderState_CurrentTABufferIdx,r7),r0	! Swap TA buffers
		mov.l	rdRenderVSyncHandler_RenderBuffers,r1
		xor	#1,r0
		mov.b	r0,@(RenderState_CurrentTABufferIdx,r7)
		shll2	r0
		mov.l	@(r0,r1),r0
		mov.l	r0,@(RenderState_CurrentTABuffer,r7)

		bsr	rdSet3DBuffer				! Start rendering
		mov.l	@(RenderState_Current3DBuffer,r7),r4


rdRenderVSyncHandler_nStartNewRender:
		lds.l	@r15+,pr
		rts
		nop

					.ALIGN	5
rdRenderVSyncHandler_Render:		.LONG	_Render
rdRenderVSyncHandler_RenderState:	.LONG	_Render+Render_State
rdRenderVSyncHandler_RenderBuffers:	.LONG	RenderBuffers
rdRenderVSyncHandler_scSetBuffer:	.LONG	_scSetBuffer



		.ALIGN	5
rdRenderDoneHandler:
		mov.l	rdRenderDoneHandler_RenderState,r1
		mov	#RENDERSTATE_DONE,r0
		rts
		mov.b	r0,@(RenderState_CurrentRenderState,r1)

					.ALIGN	2
rdRenderDoneHandler_RenderState:	.LONG	_Render+Render_State


	
!########################################################################
! Data area

			.SECTION .data

			.ALIGN	3
RenderBuffers:		.LONG	RenderBuffer0
			.LONG	RenderBuffer1

			.SECTION .bss


			.GLOBL	_Render
			.ALIGN	5
_Render:		.SPACE	Render_SIZEOF
RenderBuffer0:		.SPACE	RenderBuffer_SIZEOF
RenderBuffer1:		.SPACE	RenderBuffer_SIZEOF

			.GLOBL	_FrameCounter
			.ALIGN	2
_FrameCounter:		.SPACE	4
	
			.GLOBL	_VertexBuffers
			.ALIGN	5
_VertexBuffers:		.SPACE	5*VertexBuffer_SIZEOF
DirectPathSavedWritePtr: .SPACE	4

			.ALIGN	2
RenderDoneUserCallback:	.SPACE	4
VSyncUserCallback:	.SPACE	4
ScanlineInt1UserCallback: .SPACE 4
ScanlineInt2UserCallback: .SPACE 4

			.SECTION .data

BGPlane:		.LONG	0x90800000, 0x20800440, 0
			.SINGLE	0.0, 480.0, 0.2
			.LONG	0x7f7f7f00
			.SINGLE	0.0, 0.0, 0.2
			.LONG	0xe0007f7f
			.SINGLE	640.0, 480.0, 0.2
			.LONG	0x407f007f

RegsToSetup:		.LONG	PVR2REGS+0x00a8,		0x15d1c951	! (PVR) memory control
			.LONG	PVR2REGS+0x00a0,		0x00000020	! (unknown)
			.LONG	PVR2REGS+0x0110,		0x00093f39	! (unknown)
			.LONG	PVR2REGS+0x0098,		0x00800408	! (unknown)
			.LONG	PVR2REGS+CULLINGVALUE*4,	0x1f800000	! polygon culling (1.0f/(2**32))
			.LONG	PVR2REGS+0x0084,		0x00000000	! (unknown)
			.LONG	PVR2REGS+0x0030,		0x00000101	! (unknown)
			.LONG	PVR2REGS+0x0088,		0x30000000	! (3D) zfar very far away
			.LONG	PVR2REGS+FOGTABLECOLOR*4,	0x007f7f7f	! (3D) fog table color
			.LONG	PVR2REGS+FOGVERTEXCOLOR*4,	0x007f7f7f	! (3D) fog vertex color
			.LONG	PVR2REGS+FOGDENSITY*4,		0x0000ff07	! (3D) fog density
			.LONG	PVR2REGS+COLORCLAMPMAX*4,	0xffffffff	! (3D) color clamp max
			.LONG	PVR2REGS+COLORCLAMPMIN*4,	0x00000000	! (3D) color clamp min
			.LONG	PVR2REGS+0x0080,		0x00000007	! (unknown)
			.LONG	PVR2REGS+CHEAPSHADOWS*4,	0x00000001	! (3D) cheap shadow
			.LONG	PVR2REGS+0x007c,		0x0027df77	! (unknown)
			.LONG	PVR2REGS+COREDISABLE*4,		0x00000001	! TA reset
			.LONG	PVR2REGS+COREDISABLE*4,		0x00000000	! TA out of reset

			.LONG	PVR2REGS+0x0118,		0x00008040	! (3D) Output colour scale/add ?
			.LONG	PVR2REGS+OUTPUTCOMPRESS*4,	0x00000400	! (3D) output XY compression

			.LONG	0xa05f6884,			0x00000000	! (unknown)
			.LONG	0xa05f6930,			0x00000000	! disable all interrupt events
			.LONG	0xa05f6938,			0x00000000
			.LONG	0xa05f6900,			0xffffffff	! clear all pending int events
			.LONG	0xa05f6908,			0xffffffff

			.LONG	-1
